# -*- coding: utf-8 -*-
# @Time : 2022/3/25 14:11
# @Author : Zhan Yong
from ccdt.dataset.base_labelme.base_labelme import BaseLabelme, PathTest, Encoder, json, collections, np
from pycocotools.coco import COCO
from tqdm import *
import os
from collections import defaultdict


class Coco(BaseLabelme):
    """
    coco实现类，主要实现coco转labelme，labelme转coco，同时继承BaseLabelme
    子类就可以访问到基类（父类）的属性和方法了，它提高了代码的可扩展性和重用行。
    """

    def __init__(self, *args, **kwargs):
        self.coco_dataset = {
            "categories": [],
            "images": [],
            "annotations": [],
            "info": "",
            "licenses": [],
        }  # the complete json
        self.anns = dict()  # anns[annId]={}
        self.cats = dict()  # cats[catId] = {}
        self.imgs = dict()  # imgs[imgId] = {}
        self.imgToAnns = defaultdict(list)  # imgToAnns[imgId] = [ann]
        self.catToImgs = defaultdict(list)  # catToImgs[catId] = [imgId]
        self.imgNameToId = defaultdict(list)  # imgNameToId[name] = imgId
        self.maxAnnId = 0
        self.maxImgId = 0
        self.category_id = 0
        self.result_addImage_id = 0
        # 定义一个目录，每次labelme基类循环某一个目录的时候，追加所有数据
        # self.coco_data_infos = []

        # # 注释ID
        # self.cur_ann_id = 0
        # # 图片ID
        # self.cur_image_id = 0
        # # 图片对象列表
        # self.images = list()
        # # 注释对象列表
        # self.annotations = list()
        # # 类别对象列表
        # self.categories = list()
        # # 所有标注属性对象列表
        self.coco_shapes = list()
        # # 目标检测类别对象列表
        self.categories_name = list()

        # 自定义coco文件名称
        self.coco_file_name = 'coco.json'
        self.coco_file = kwargs.get('coco_file')
        self.coco = None
        if self.coco_file:
            self.coco = COCO(self.coco_file)  # 加载coco文件数据
        super().__init__(*args, **kwargs)

    def has_image(self, image_name):
        """
        图片文件id查询
        :param image_name:
        :return:
        """
        img_id = self.imgNameToId.get(image_name, None)
        return img_id is not None

    def add_image(self, file_name: str, width: int, height: int, id: int = None):
        """
        追加图片描述属性，返回图片image_id
        :param file_name:
        :param width:
        :param height:
        :param id:
        :return:
        """
        if self.has_image(file_name):
            print(f"{file_name}图片已存在")
            return
        if not id:
            self.maxImgId += 1
            id = self.maxImgId
        image = {
            "id": id,
            "width": width,
            "height": height,
            "file_name": file_name
        }
        self.coco_dataset["images"].append(image)
        self.imgs[id] = image
        self.imgNameToId[file_name] = id
        return id

    def get_bbox(self, bbox, image_w_h, file_path):
        """
        实现矩形框转换计算，得到：左上角的坐标点+宽+高，即[x，y，宽，高]
        :param bbox:
        :param image_w_h:
        :return:
        """
        # points = np.array([1,2,3,4])
        # points = np.array([])
        # points[0, 0] = '赋值'
        # print(image_w_h)
        # shape['points'] == bbox，值为：[[453.59124755859375, 81.40898895263672], [654.3768920898438, 279.54791259765625]]
        if len(bbox) == 2:
            x1 = bbox[0][0]
            y1 = bbox[0][1]
            x2 = bbox[1][0]
            y2 = bbox[1][1]
            # 只针对坐标点越界的矩形进行处理
            if x1 < 0 or y1 < 0 or x2 < 0 or y2 < 0 or x1 > image_w_h[0] or y1 > image_w_h[1] or x2 > image_w_h[
                0] or y2 > \
                    image_w_h[1]:
                print(f'标注的矩形框的坐标点已经超越图像边界:{file_path}')
                clamp_x1 = np.clip(x1, 0, image_w_h[0])
                clamp_y1 = np.clip(y1, 0, image_w_h[1])
                clamp_x2 = np.clip(x2, 0, image_w_h[0])
                clamp_y2 = np.clip(y2, 0, image_w_h[1])
                # 替换
                bbox[0][0] = clamp_x1
                bbox[0][1] = clamp_y1
                bbox[1][0] = clamp_x2
                bbox[1][1] = clamp_y2
            points = np.array(bbox)
            point_min, point_max = points.min(axis=0), points.max(axis=0)
            w, h = point_max - point_min
            return [point_min[0], point_min[1], w, h]

    def get_area(self, bbox):
        """
        计算矩形框的面积
        :param bbox:
        :return:
        """
        points = np.array(bbox)
        point_min, point_max = points.min(axis=0), points.max(axis=0)
        w, h = point_max - point_min
        return w * h
        # x = bbox[::2]
        # y = bbox[1::2]
        # return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))

    def add_annotation(self, image_id: int, category_id: int, segmentation: list, bbox: list, id: int = None,
                       image_w_h: tuple = (), file_name: str = ''):
        """
        追加coco标注属性
        :param image_id:
        :param category_id:
        :param segmentation:
        :param bbox:
        :param id:
        :param image_w_h:
        :param file_name:
        :return:
        """
        if id is not None and self.anns.get(id, None) is not None:
            print("标签已经存在")
            return
        if not id:
            self.maxAnnId += 1
            id = self.maxAnnId
        ann = {
            "id": id,
            "iscrowd": 0,
            "image_id": image_id,
            "category_id": category_id,
            "segmentation": [segmentation],
            "area": self.get_area(bbox),
            "bbox": self.get_bbox(bbox, image_w_h, file_name),
        }
        self.coco_dataset["annotations"].append(ann)
        self.anns[id] = ann
        self.imgToAnns[image_id].append(ann)
        self.catToImgs[category_id].append(image_id)
        return id

    def add_category(self, id: int, name: str, color: list, supercategory: str = ""):
        """
        追加类别属性
        :param id:
        :param name:
        :param color:
        :param supercategory:
        """
        cat = {
            "id": id,
            "name": name,
            "color": color,
            "supercategory": supercategory,
        }
        self.cats[id] = cat
        self.coco_dataset["categories"].append(cat)

    def self2labelme(self):
        """
        coco转labelme实现
        """
        # 如果self.coco为空，则实现labelme转coco
        if self.coco is None:
            return
        # 获取每一张图片的id
        img_ids = self.coco.getImgIds()
        # 循环遍历每一个id
        i = 0
        for img_id in tqdm(img_ids):
            img_info = self.coco.loadImgs(img_id)[0]
            obj_dir = PathTest.initialize(img_info['file_name'])
            relative_path = os.path.join('..', '00.images', img_info['file_name'])
            # 通过注释id寻找到同一帧图片下的多个注释属性
            labelme_data = self.single_coco2labelme(img_id)
            labelme_data['imageHeight'] = img_info['height']
            labelme_data['imageWidth'] = img_info['width']
            labelme_data['imagePath'] = relative_path
            data_info = dict(image_dir=self.images_dir,
                             image_file=img_info['file_name'],
                             labelme_dir=self.labelme_dir,
                             labelme_file=obj_dir.stem + '.json',
                             labelme_info=labelme_data,
                             data_type=self.data_type,
                             input_dir=self.input_dir,
                             output_dir=self.output_dir,
                             only_annotation=self.only_annotation)
            self.data_infos.append(data_info)
            i += 1
        print("coco转labelme结束,一共%d" % i)

    def single_coco2labelme(self, img_id):
        """
        实现每一个coco注释合成labelme注释
        @param img_id:
        @return:
        """
        ann_ids = self.coco.getAnnIds(imgIds=img_id)
        labelme_data = dict(
            version='4.5.9',
            flags={},
            shapes=[],
            imagePath=None,
            imageData=None,
            imageHeight=None,
            imageWidth=None
        )
        if ann_ids:
            shapes = []
            anns = self.coco.loadAnns(ann_ids)
            for ann in anns:
                # 取到同一张图片的多个注释属性
                category_id = ann['category_id']
                # 获取坐标框
                bbox = ann['bbox']
                # 坐标切割
                bbox = [bbox[0:2], bbox[2:4]]
                # 左上角的坐标(x,y)右上角的坐标(x,y+h)左下角的坐标(x+w,y)右下角的坐标(x+w,y+h)
                points = [bbox[0], [bbox[0][0] + bbox[1][0], bbox[0][1] + bbox[1][1]]]
                # 通过类别id获取类别名称
                cats = self.coco.loadCats(category_id)[0]
                # 取到类别名称
                name = cats['name']
                shape = {"label": name, "points": points, "group_id": None, "shape_type": "rectangle", "flags": {}}
                shapes.append(shape)
            labelme_data['shapes'] = shapes
        return labelme_data

    def self2coco(self):
        """
        labelme转coco实现
        """
        # 如果self.coco不为空，则实现coco转labelme
        if self.coco:
            return
        self.get_data_paths()
        self.load_labelme()
        # try:
        self.update_property()
        # except Exception as e:
            # print(e)
        print('后面不执行了吗')
        if self.data_infos:
            print('还是说这里就问题了')
            # self.coco_data_infos.extend(self.data_infos)
            # 定义coco数据格式
            # coco_data = dict(
            #     images=self.images,
            #     annotations=self.annotations,
            #     categories=self.categories,
            # )
            # 递归的目录转coco，有多少个文件夹就会被类实列多少次。递归多少次文件夹，这里就会被循环多少次，不管循环多少次，这里需要生成一个coco文件
            for labelme_info in self.data_infos:
                print(labelme_info)
                # self.coco_data_infos.append(labelme_info)
                # 针对背景类直接跳过、有json文件没有图片文件直接跳过，不写入图片属性
                if labelme_info.get('labelme_info') is None or labelme_info.get('labelme_info').get('shapes') == [] \
                        or labelme_info.get('image_file') is None:
                    continue
                # 1、添加coco类别

                # 2、追加每张图片文件属性
                self.result_addImage_id = self.add_image(labelme_info.get('image_file'),
                                                         labelme_info.get('labelme_info').get('imageWidth'),
                                                         labelme_info.get('labelme_info').get('imageHeight'), None)
                w = labelme_info.get('labelme_info').get('imageWidth')
                h = labelme_info.get('labelme_info').get('imageHeight')
                size = (w, h)  # 把宽高定义成元组
                file_path = os.path.join(labelme_info.get('input_dir'), labelme_info.get('image_dir'),
                                         labelme_info.get('image_file'))
                # print(type(size))
                # self.cur_image_id += 1
                # images = dict(
                #     file_name=labelme_info['image_file'],
                #     height=labelme_info['labelme_info']['imageHeight'],
                #     width=labelme_info['labelme_info']['imageWidth'],
                #     id=self.cur_image_id,
                # )
                # self.images.append(images)
                # self.coco_shapes = labelme_info['labelme_info']['shapes']
                # 标签属性处理
                one_img_ann_list = self.labelme_shapes2coco_ann(labelme_info)
                for ann in one_img_ann_list:
                    print(ann)
                    category_id = 0
                    # 如果标注的标签不存在，就category_id加1
                    if ann['label'] not in self.categories_name:
                        self.category_id += 1
                        self.add_category(self.category_id, str(ann['label']), [], '')
                        self.categories_name.append(ann['label'])
                    # 每一个shape元素中动态查询，
                    for category in self.coco_dataset.get('categories'):
                        print(category)
                        if category.get('name') == ann['label']:
                            category_id = category.get('id')
                            # print(category_id)
                    # 3、追加标注属性,核心是self.category_id要动态追加，必须要在,self.coco_dataset.get('categories')取
                    self.add_annotation(self.result_addImage_id, category_id, ann.get('segmentation'),
                                        ann.get('points'), None, size, file_path)


                # print(self.coco_dataset)
                # self.annotations.extend(one_img_ann_list)
            # for category_id, category_name in enumerate(self.categories_name, 1):
            #     category = dict(
            #         supercategory='object',
            #         id=category_id,
            #         name=category_name,
            #     )
            #     self.categories.append(category)
            # obj_dir = PathTest.initialize(self.images_dir)
            # self.save_labelme()
            # out_put_coco_file = obj_dir.joinpath(self.output_dir, obj_dir.parent, self.coco_file_name)
            # coco文件的名称是写死的

            # 创建输出目录文件夹
            os.makedirs(os.path.join(self.output_dir), exist_ok=True)
            out_put_coco_file = os.path.join(self.output_dir, self.coco_file_name)
            with open(out_put_coco_file, 'w', encoding='utf-8') as coco_fp:
                json.dump(self.coco_dataset, coco_fp, ensure_ascii=False, indent=2, cls=Encoder)
        print('labelme转coco结束，类别打印时间久，可以手动结束程序')
        print('注意：labelme转coco结束，只能是单级目录下的labelme数据集进行转换')
        return
        # except Exception as e:
        #     print(e)

    @staticmethod
    def labelme_shapes2coco_ann(labelme_info):
        """
        把labeme标签转成coco标签
        针对shape元素二次封装，把group_id做为key，矩形框和关键点作为值。{'group_id': [][], 'b': 2, 'b': '3'}
        找出所有shape元素中相同group_id组的shape元素，合并为一个列表，即一个矩形框对应多个关键点标注。使用：structure_shapes = defaultdict(list)
        @return:
        """
        # tinydict = {}
        # shapes_type = collections.defaultdict(list)
        # from collections import defaultdict
        # 封装新的关键点与矩形框组合列表。即group_id的值作为键，shape元素作为值，构成新的字典集合
        structure_shapes = defaultdict(list)
        # i = 0
        # 筛选group_id相同的shape元素，为一个列表集合中，参考网址关键词:字典设置默认值，https://blog.csdn.net/qq_45128719/article/details/125392954
        for shape in labelme_info.get('labelme_info').get('shapes'):
            print(shape)
            # rectangle(矩形框),point(关键点),polygon(多边形)，line(线)，linestrip(线带，语义分割)
            if shape['shape_type'] == 'rectangle' or shape['shape_type'] == 'point' or \
                    shape['shape_type'] == 'polygon' or shape['shape_type'] == 'line' or \
                    shape['shape_type'] == 'linestrip':
                # ====================================================
                # # # clamp 标注的矩形框，超越图片宽高边界的，需要对矩形框的坐标点做一次，夹紧计算。
                # # 针对，左上角的坐标点与右下角的坐标点的，计算公式：如下
                # # np.clip(x1, 0, 图像宽)，
                # # np.clip(x1, 0, 图像宽)，
                # # np.clip(y1, 0, 图像高)，
                # # np.clip(x2, 0, 图像宽)，
                # # np.clip(y1, 0, 图像高)，
                # # 最后，在夹紧计算出4个坐标点后，labelme转coco时从新计算[x，y，宽，高]
                # w = labelme_info.get('labelme_info').get('imageWidth')
                # h = labelme_info.get('labelme_info').get('imageHeight')
                # size = (w, h)  # 把宽高定义成元组
                # x1 = shape.get('points')[0][0]
                # y1 = shape.get('points')[0][1]
                # x2 = shape.get('points')[1][0]
                # y2 = shape.get('points')[1][1]
                # 逻辑判断坐标点是否超越图像边界
                # if x1 < 0 or y1 < 0 or x2 < 0 or y2 < 0 or x1 > size[0] or y1 > size[1] or x2 > size[0] or y2 > size[1]:
                #     file_path = os.path.join(labelme_info.get('input_dir'), labelme_info.get('image_dir'), labelme_info.get('image_file'))
                #     print('标注的矩形框的坐标点已经超越图像边界:' + file_path)
                # clamp_x1 = np.clip(x1, 0, size[0])
                # clamp_y1 = np.clip(y1, 0, size[1])
                # clamp_x2 = np.clip(x2, 0, size[0])
                # clamp_y2 = np.clip(y2, 0, size[1])
                # ====================================================
                # 针对客户标注的错误，导致转coco时，矩形框出错，该程序不做逻辑处理，给使用者自己过滤，保障shape不少即可
                structure_shapes[shape['group_id']].append(shape)
                # # group_id的值为None，表示当前矩形框没有关键点，需要单独一个shape元素
                # if not shape.get('group_id') and shape.get('group_id') != 0:
                #     id_value = shape.get('group_id')
                #     print(f'group_id的值为空={id_value}')
                #     # print('group_id的值为空' +f{shape.get('group_id')} )
                #     # return self.coco_shapes
                # else:
                #     id_value = shape.get('group_id')
                #     print(f'group_id的值,不为空={id_value}')
                #     # 这里的转换存在一个问题，如果某一个shape打了group_id的值，但没有打关键点，同时标签类别名称和其它有关键点的矩形框相同时。
                #     # group_id值相同会被分一组，如果认为标注出错，分组就会出错，如何避免。会导致转coco的矩形框变少，并且标错位置。
                #     # 针对上述错误的矫正逻辑思路，查询每一个shape字典中的group_id，并且和整个shapes元素进行比较，相同值的放一个列表集合。
                #     structure_shapes[shape['group_id']].append(shape)
        # print(structure_shapes)
        # one_img_ann_list = list()
        rebuild_labelme_infos = []  # 重组一张图片多个shape元素的集合
        for structure_shape, value in structure_shapes.items():
            if not structure_shape and structure_shape != 0:
                # print('不是关键点与矩形框的分组数据')
                rebuild_labelme_infos.extend(value)  # 直接列表相加扩展
            else:
                # 处理标注的shape有group_id的元素，只要group_id有值，就追加segmentation
                rebuild_shape = {}  # 重组shape字典，追加关键点属性，把一个组的多个shape元素，合并成一个shape。针对关键点与shape的组合
                segmentation = []  # 关键点列表定义
                for shape in value:
                    # 如果标注shape是关键点，就追加关键点元素列表
                    if shape['shape_type'] == 'point':
                        segmentation.extend(shape['points'][0])
                    else:
                        rebuild_shape.update(shape)
                # 重组shape属性，并追加关键点列表
                rebuild_shape.update({'segmentation': segmentation})
                rebuild_labelme_infos.append(rebuild_shape)
                # print('组合结束')
        return rebuild_labelme_infos
        # # 添加标签类别名称
        # if structure_shape['label'] not in self.categories_name:
        #     self.categories_name.append(structure_shape['label'])
        # # 处理shape属性
        # ann_list = self.rectangle_shapes2coco(structure_shape)
        # one_img_ann_list.extend(ann_list)

        # for index, shape in enumerate(rebuild_labelme_infos):
        #     # if shape['group_id'] == shape['group_id']:
        #     #     print(shape)
        #         # tinydict.fromkeys(shape['group_id'][shape])
        #
        #     # # 添加标签类别名称
        #     if shape['label'] not in self.categories_name:
        #         self.categories_name.append(shape['label'])
        #         # if shape_type == 'rectangle':
        #         ann = self.get_default_ann()
        #         category_id = index + 1
        #         ann['category_id'] = category_id
        #         points = np.array(shape['points'])
        #         point_min, point_max = points.min(axis=0), points.max(axis=0)
        #         w, h = point_max - point_min
        #         ann['bbox'] = [point_min[0], point_min[1], w, h]
        #         ann['area'] = w * h
        #         ann['segmentation'].extend(shape['segmentation'])
        #         self.annotations.append(ann)

        # ann_list = self.rectangle_shapes2coco(shape)

        # # 处理shape属性
        # ann_list = self.rectangle_shapes2coco(shape)
        # one_img_ann_list.extend(ann_list)

        # if shape['shape_type'] == 'rectangle':
        #     if shape['label'] not in self.categories_name:
        #         self.categories_name.append(shape['label'])
        #     shapes_type[shape['shape_type']].append(shape)
        # elif shape['shape_type'] == 'polygon':
        #     continue
        # elif shape['shape_type'] == 'line':
        #     continue
        # elif shape['shape_type'] == 'linestrip':
        #     continue
        # elif shape['shape_type'] == 'circle':
        #     continue
        # elif shape['shape_type'] == 'point':
        #     if shape['label'] not in self.categories_name:
        #         self.categories_name.append(shape['label'])
        #     shapes_type[shape['shape_type']].append(shape)
        # one_img_ann_list = list()
        # 目前来看，68个关键点，每张图片，只有一个矩形框有关键点。不是所有矩形框都有关键点。
        # for shape_type, shapes_type_ann in shapes_type.items():
        #     if shape_type == 'rectangle':
        #         ann_list = self.rectangle_shapes2coco(shapes_type.get(shape_type))
        #         one_img_ann_list.extend(ann_list)
        #     elif shape_type == 'polygon':
        #         continue
        #     elif shape_type == 'line':
        #         continue
        #     elif shape_type == 'linestrip':
        #         continue
        #     elif shape_type == 'circle':
        #         continue
        #     elif shape_type == 'point':
        #         # 关键点转换到coco文件中的segmentation中
        #         ann_list = self.point_shapes2coco(shapes_type.get(shape_type))
        #         one_img_ann_list.extend(ann_list)
        # return one_img_ann_list

    @staticmethod
    def save_coco():
        """
        coco输出，想要把多个文件夹，的文件一次性输出到一个文件，并且有编号顺序，必须是一次遍历多个目录下的所有文件写coco内存数据，而不是每个目录分成一个对象。
        """
        print('coco保存')
    # def rectangle_shapes2coco(self, shapes):
    #     """
    #     实现矩形框转换计算
    #     @param shapes:
    #     @return:
    #     """
    #     ann_list = []
    #     for shape in shapes:
    #         ann = self.get_default_ann()
    #         category_id = shapes.index(shape) + 1
    #         ann['category_id'] = category_id
    #         points = np.array(shape['points'])
    #         point_min, point_max = points.min(axis=0), points.max(axis=0)
    #         w, h = point_max - point_min
    #         ann['bbox'] = [point_min[0], point_min[1], w, h]
    #         ann['area'] = w * h
    #         ann_list.append(ann)
    #     return ann_list

    # def get_default_ann(self):
    #     """
    #     标签属性定义，并id自增
    #     关键点存放格式如下所示
    #     "segmentation": [
    #             [
    #                 491.0,
    #                 637.0,
    #                 252.0,
    #                 617.0,
    #                 234.0,
    #                 480.0,
    #                 473.0,
    #                 500.0
    #             ]
    #         ],
    #     一张图片，只有一个矩形框，多个关键点比较好弄，如果是多个矩形框，有的矩形框有关键点，有的矩形框没有关键点。
    #     也就是说，关键点和矩形框的对应关系，目前无法确认。（通过组group_id的关系来确认，关键点归属于哪一个矩形框）
    #     @return:
    #     """
    #     self.cur_ann_id += 1
    #     annotation = dict(
    #         segmentation=[],
    #         area=0,
    #         iscrowd=0,  # 做为group_id，关联字段
    #         image_id=self.cur_image_id,
    #         bbox=[],
    #         category_id=0,
    #         id=self.cur_ann_id,
    #     )
    #     return annotation

    # def point_shapes2coco(self, shapes):
    #     """
    #     关键点追加，一个shape矩形框下，就是68个关键点
    #     :param shapes:
    #     :return:
    #     """
    #     ann_list = []
    #     ann = self.get_default_ann()
    #     for shape in shapes:
    #         category_id = shapes.index(shape) + 1
    #         ann['category_id'] = category_id
    #         # points = np.array(shape['points'])
    #         # point_min, point_max = points.min(axis=0), points.max(axis=0)
    #         # w, h = point_max - point_min
    #         # ann['bbox'] = [point_min[0], point_min[1], w, h]
    #         # 算的是关键点连接成线的面积
    #         # ann['area'] = w * h
    #         x = shape.get('points')[0][0]
    #         y = shape.get('points')[0][1]
    #         ann['segmentation'].append(x)  # 追加关键点，待完善
    #         ann['segmentation'].append(y)
    #     ann_list.append(ann)
    #     return ann_list



# -*- coding: utf-8 -*-
# @Time : 2022/3/25 13:54
# @Author : Zhan Yong
import prettytable as pt


class LabelInfo(object):
    """
    创建可迭代对象，操作labelme的json文件中shapes列表中包含的字典对象属性。最节约内存
    """

    def __init__(self, shapes_data):
        self.shapes_data = shapes_data
        self.number = 0
        self.shapes_data_info = dict()

    def __iter__(self):
        return self

    def __next__(self):
        """
        每次迭代出想要的内容，又能抛出异常终止迭代，raise StopIteration
        #  https://www.cnblogs.com/xifengmo/p/11029391.html
        :return:
        """
        # cursor = 0
        # while cursor < len(self.shapes_data):
        #     yield self.shapes_data.items[cursor]
        #     cursor += 1

        if self.number < self.shapes_data.__len__():
            for label_key, label_infos in self.shapes_data.items():
                self.label_info(label_key, label_infos)
                self.number += 1
            return self.shapes_data_info
        else:
            raise StopIteration  # 抛出异常停止遍历

    def label_info(self, label_key, label_info):
        """
        封装同一个标签类别对应的相关属性为字典，方便打印输出
        封装后数据结构：{{'person': {'shape_type': ['rectangle'], 'flags': ['狗狗', '活', '安全帽', '工装'], 'group_id': [None]}}
        :param label_key: 标签类别
        :param label_info: 标签类别对应相关属性
        """
        label_shape_flag_group = dict()
        label_shape_flag_group[label_key] = {'shape_type': [], 'flags': [], 'group_id': []}
        shape_type = list()
        group_id = list()
        flags = list()
        for label in label_info:
            shape_type += label[0]
            flags += label[1]
            group_id += label[2]
            label_shape_flag_group[label_key]['shape_type'] = list(set(shape_type))
            label_shape_flag_group[label_key]['group_id'] = list(set(group_id))
        label_shape_flag_group[label_key]['flags'] = list(set(self.flags_info(flags)))
        return self.shapes_data_info.update(label_shape_flag_group)

    def flags_info(self, flags):
        """
        封装类别标签属性flags，把同一个标签类别下flags属性值为True的键封装成列表返回
        封装前数据结构：<class 'list'>: [{'活': False}, {'狗狗': True}, {'工装': False}, {'安全帽': True}]
        封装后数据结构：['狗狗','工装']
        :param flags:
        :return:
        """
        fg = list()
        for flag in flags:
            for key, value in flag.items():
                if flag[key] is True:
                    fg.append(key)
        return fg

    def __repr__(self):
        # property_tb = pt.PrettyTable()
        # property_tb.field_names = ['label_name', 'shape_type_name', 'flags_name']
        property_tb = pt.PrettyTable(['label_name', 'shape_type_name', 'flags_name'])
        for key, value in self.shapes_data_info.items():
            property_tb.add_row([key, value['shape_type'], value['flags']])
            # property_tb.add_row([key])
            # property_tb.add_row([value['flags']])
            # property_tb.add_row([value['shape_type']])

        # print(property_tb.get_string(title='title'))
        return str(property_tb.get_string(title='title'))
        # return str(property_tb)



# -*- coding: utf-8 -*-
# @Time : 2022/3/25 14:05
# @Author : Zhan Yong
from tqdm import tqdm
from copy import deepcopy
import collections
import cv2
import numpy as np

from .label_infos import *
from .async_io_task import *
import time
import asyncio
import os


class BaseLabelme(object):
    """
    labelme数据集处理基类，实现数据打印、标注区域抠图、类别及flags名称重命名、根据labelme属性筛选、标注可视化
    """

    def __init__(self, *args, **kwargs):
        self.shape = None
        self.only_annotation = kwargs.get('only_annotation')
        self.labelme_dir = kwargs.get('labelme_dir')
        self.images_dir = kwargs.get('images_dir')
        self.file_formats = kwargs.get('file_formats')
        self.data_type = kwargs.get('data_type')
        self.input_dir = kwargs.get('input_dir')
        self.output_dir = kwargs.get('output_dir')
        self.function = kwargs.get('function')
        self.select_empty = kwargs.get('select_empty')
        self.num_labelme = 0  # json文件数量
        self.num_images = 0  # 图片文件数量
        self.num_background = 0  # 背景图片文件数量
        self.label2label_info = collections.defaultdict(list)  # 标注属性多项集
        self.data_paths = list()
        self.data_infos = list()
        self.filter_data = dict()
        from ccdt import Coco
        if type(self) == BaseLabelme:
            self.get_data_paths()
            self.load_labelme()
        elif type(self) == Coco:
            self.self2labelme()  # coco转labelme
            self.self2coco()  # labelme转coco
        self.update_property()  # 更新属性，用于信息打印、类别筛选
        # if self.function == 'merge':
        #     # 这里有一个错误，指定文件夹做为类别判断，如果文件夹变动会出错
        #     if self.images_dir not in self.label2label_info.keys():
        #         print(self.label2label_info.keys())
        #         assert False, '{} 类别合并功能发现数据集类别不对，请核对筛选类别'.format(self.images_dir)

    def __repr__(self):
        """
        数据集表格打印实现
        :return:
        """
        num_tb = pt.PrettyTable()
        num_tb.field_names = ['num_images', 'num_labelme', 'num_background']
        num_tb.add_row(
            [self.num_images, self.num_labelme, self.num_background])
        return str(num_tb)

    def __call__(self, *args, **kwargs):
        """
        把该类的实列变为可调用对象
        :param args:
        :param kwargs:其中filter_empty=False保留背景类，为True不保留背景类。
        only_empty=False为不保留背景类，为True只筛选背景类。
        name_classes为标签类别
        filter_flags为标签类别属性
        shape_type为标注形状
        """
        self.filter_label = kwargs.get('filter_label')
        self.filter_flags = kwargs.get('filter_flags')
        self.only_select_shapes = kwargs.get('only_select_shapes')
        self.filter_shape_type = kwargs.get('filter_shape_type')
        self.filter_combin = kwargs.get('filter_combin')
        self.shapes_attribute = kwargs.get('shapes_attribute')
        self.only_select_empty = kwargs.get('only_select_empty')
        self.auto_filter()
        return self

    def get_data_paths(self):
        """
        封装数据路径方法实现，把图片路径、图片名称、json路径、json名称，一一对应并封装成一个字典
        :return:
        """
        from ccdt import PathTest
        # 针对不同的功能实现，需要有不同路径组合，需要不同的判断条件
        if self.function == 'merge':  # 合并功能
            self.data_paths = PathTest.get_merge_paths(self.images_dir, self.file_formats, self.input_dir,
                                                       self.output_dir, self.data_type, self.only_annotation, )
        # 打印功能、筛选功能、重命名功能、抠图功能、labelme转coco功能
        if self.function == 'filter' or self.function == 'print' \
                or self.function == 'rename' or self.function == 'matting' or self.function == 'convert' \
                or self.function == 'delete':
            self.data_paths = PathTest.get_filter_paths(self.images_dir, self.file_formats, self.input_dir,
                                                        self.output_dir, self.data_type, self.only_annotation,
                                                        self.labelme_dir)
        return self.data_paths

    def load_labelme(self):
        """
        封装labelme的json文件数据，把整个json对象封装到labelme_info字典中
        :return:
        """
        print("加载labelme数据进度:")
        # ==============================异步读取数据到内存操作==================================================
        # async_time = time.time()
        # loop = asyncio.get_event_loop()
        # # future = asyncio.ensure_future([AsyncIoTask.json_load(data_path) for data_path in tqdm(self.data_paths)])
        # # loop.run_until_complete(future)
        # # print(future.result())
        # tasks = [AsyncIoTask.json_load(data_path) for data_path in tqdm(self.data_paths)]
        # loop.run_until_complete(asyncio.wait(tasks))
        # loop.close()
        # print('异步耗时')
        # print(time.time() - async_time)
        # ==============================异步读取数据到内存操作===================================================

        # ==============================同步读取数据到内存操作===================================================
        # start_time = time.time()
        for data_path in tqdm(self.data_paths):
            # 组合加载json文件的路径
            labelme_path = os.path.join(data_path['input_dir'], data_path['labelme_dir'], data_path['labelme_file'])
            try:
                with open(labelme_path, 'r', encoding='UTF-8') as labelme_fp:
                    data_path['labelme_info'] = json.load(labelme_fp)
                    if data_path['labelme_info']['imageData'] is not None:
                        data_path['labelme_info']['imageData'] = None
                    if not data_path['labelme_info']['shapes']:
                        data_path['background'] = True
            except Exception as e:
                # 如果没有json文件，读取就跳过，并设置为背景
                if 'No such file or directory' in e.args:
                    data_path['background'] = True
                    data_path['labelme_file'] = None
                else:  # 如果是其它情况错误（内容为空、格式错误），就删除json文件并打印错误信息
                    print(e)
                    print(labelme_path)
                    os.remove(labelme_path)
                data_path['background'] = True
            self.data_infos.append(data_path)
        # print(time.time() - start_time)
        # if async_time < start_time:
        #     print('读取数据到内存操作,异步耗时短')
        # if async_time > start_time:
        #     print('读取数据到内存操作,同步耗时短')
        return self.data_infos
        # ==============================同步读取数据到内存操作===================================================

    def self2labelme(self):
        """
        coco转labelme
        coco子类没有实现,就报错
        """
        raise NotImplementedError

    def self2coco(self):
        """
        labelme转coco
        coco子类没有实现,就报错
        """
        raise NotImplementedError

    @property
    def parser(self):
        """
        封装label、shape_type、flags、group_id属性，即把同一个标签下的全部属性封装列表里
        :return:
        """
        label = list()
        shape_type = list()
        flags = list()
        group_id = list()
        if 'label' in self.shape:
            if self.shape['label'] not in label:
                label.append(self.shape['label'])
        if 'shape_type' in self.shape:
            if self.shape['shape_type'] not in shape_type:
                shape_type.append(self.shape['shape_type'])
        if 'group_id' in self.shape:
            if self.shape['group_id'] not in group_id:
                group_id.append(self.shape['group_id'])
        if 'flags' in self.shape:
            if self.shape['flags'] not in flags:
                flags.append(self.shape['flags'])
        return label, shape_type, flags, group_id

    def update_property(self):
        """
        更新属性实现方法，包含统计背景类、类别标签对应属性合并
        """
        num_shapes = 0  # 计数shape为空
        num_labelme = 0  # 计数labelme_info
        num_images = 0  # 计数image_file
        num_background = 0  # 计数labelme_info为空并且image_file不为空.有图片没有标注的算背景，有json文件就算背景不对
        for data_info in self.data_infos:
            if data_info.get('image_file') and data_info.get('labelme_info') is None:
                num_background += 1
            if data_info.get('image_file'):
                num_images += 1
            if data_info.get('image_file') and data_info.get('labelme_info'):
                num_labelme += 1
                if not data_info.get('labelme_info').get('shapes'):
                    num_shapes += 1
                for shape in data_info.get('labelme_info').get('shapes'):
                    # 为类动态添加属性
                    self.shape = shape
                    label, shape_type, flags, group_id = self.parser
                    self.label2label_info[label[0]].append([shape_type, flags, group_id])
        self.num_images = num_images
        self.num_labelme = num_labelme
        self.num_background = num_background + num_shapes
        print('这里没有错误')
        return self

    def crop_objs(self, min_pixel=10, ):
        """
        截取图像功能实现，一张图片画框多少，就扣多少，不管是否重叠
        标注形状类型，比如rectangle(长方形)、circle(圆)、polygon(多边形)、line(线)
        :param out_dir: 保存截取图像路径
        :param min_pixel:保存截取图片最小像素设置
        """
        print("截图图片进度:")
        for data_info in tqdm(self.data_infos):
            assert data_info['image_file'], '传入的图片路径为空，不能进行图片截取：{}'.format(data_info['labelme_file'])
            if data_info['labelme_file'] is None or data_info['labelme_info']['shapes'] == []:
                continue
            num_obj = 0
            for shape in data_info['labelme_info']['shapes']:
                # 组合保存抠图类别目录（Z:/4.my_work/9.zy/matting/00/00.images/call）
                save_img_dir = os.path.join(data_info['output_dir'], data_info['image_dir'], shape['label'])
                os.makedirs(save_img_dir, exist_ok=True)
                # 组合图片路径
                image_path = os.path.join(data_info['input_dir'], data_info['image_dir'], data_info['image_file'])
                # 组合抠图图片名称
                from ccdt import PathTest
                obj_img_file = PathTest.initialize(image_path)
                image = cv2.imdecode(np.fromfile(image_path, dtype=np.uint8), 1)
                num_obj += 1
                crop_file_name = obj_img_file.parent.parent.stem + '_' + obj_img_file.stem + '_{:0>6d}'.format(
                    num_obj) + obj_img_file.suffix
                # 组合抠图图片存储路径
                crop_file_path = os.path.join(save_img_dir, crop_file_name)
                crop_obj = self.crop_rectangle(image, shape)
                if crop_obj.size == 0:
                    print("当前文件标注存在异常，路径如下所示:")
                    print(crop_file_path)
                # 默认像素小于10，就不进行截取，可以自动设置
                if crop_obj.shape[0] * crop_obj.shape[1] > min_pixel:
                    cv2.imencode(obj_img_file.suffix, crop_obj)[1].tofile(crop_file_path)

    @staticmethod
    def crop_rectangle(image, shape):
        """
        长方形截取计算过程
        :param image: 图像
        :param shape: 形状坐标
        :return:
        """
        h = image.shape[0]
        w = image.shape[1]
        # 把从任意角度标注兼容计算
        points = np.array(shape['points'])
        point_min, point_max = points.min(axis=0), points.max(axis=0)
        x1, y1 = int(max(0, min(point_min[0], w))), int(max(0, min(point_min[1], h)))
        x2, y2 = int(max(0, min(point_max[0], w))), int(max(0, min(point_max[1], h)))
        # y1:y2 x1:x2,针对负数框在图片的外面时截取不到,正常标注不会超出图片面积范围。max(0, min(x, img_info['width'])把负数变成0。np.clip(point_min)
        crop_obj = image[y1:y2, x1:x2]
        return crop_obj

    def rename(self, rename):
        """
        重命名功能实现，包含label和flags
        :param rename:
        """
        for data_info in self.data_infos:
            if data_info['labelme_info'] is not None:
                for shape in data_info['labelme_info']['shapes']:
                    if rename.get('label') is None and rename.get('flags') is None:
                        assert False, '输入的（--rename-label）重命名属性值为空'.format(rename)
                    if rename.get('label') and rename.get('flags'):
                        self.label_rename(shape, rename)
                        self.flags_rename(shape, rename)
                    if rename.get('flags'):
                        self.flags_rename(shape, rename)
                    if rename.get('label'):
                        self.label_rename(shape, rename)

    @staticmethod
    def label_rename(shape, rename):
        """
        重命名label
        :param shape:
        :param rename:
        """
        if shape['label'] in rename.get('label'):  # 修改标签类别名称
            shape['label'] = rename.get('label').get(shape['label'])

    @staticmethod
    def flags_rename(shape, rename):
        """
        重命名flags
        :param shape:
        :param rename:
        """
        # 判断一个列表的元素是否在另一个列表中
        if set(rename.get('flags').keys()).issubset(shape['flags'].keys()):
            for rename_key in rename.get('flags').keys():  # 修改标签类别属性名称
                shape['flags'][rename.get('flags')[rename_key]] = shape['flags'].pop(rename_key)

    def save_labelme(self):
        """
        根据类别保存labelme。coco转labelme需要现在内存中与datainfos关联上，然后用该方法存储到磁盘上。
        """
        print('处理数据保存处理进度:')
        # ==============================异步写内存数据到磁盘操作==================================================
        loop = asyncio.get_event_loop()
        async_time = time.time()
        if self.filter_data:
            filter_data_info = []
            for value in self.filter_data.values():
                for data in value:
                    filter_data_info.append(data)
            tasks = [AsyncIoTask.json_dump(value) for value in tqdm(filter_data_info)]
        else:
            tasks = [AsyncIoTask.json_dump(data_info) for data_info in tqdm(self.data_infos)]
        if not tasks:
            print('筛选后的保存数据为空，跳过本次处理')
            return
        loop.run_until_complete(asyncio.wait(tasks))
        print('异步耗时')
        print(time.time() - async_time)
        # ==============================异步写内存数据到磁盘操作===================================================

        # ==============================同步写内存数据到磁盘操作=========================================================
        # start_time = time.time()
        # for data_info in tqdm(self.data_infos):
        #     if data_info['image_file']:
        #         image_path = PathOperate.make_up_path(data_info['image_dir'], data_info['image_file'], out_dir, None,
        #                                               data_info['data_type'])
        #         copy_path = PathOperate.make_up_path(data_info['image_dir'], data_info['image_file'], None, None,
        #                                              data_info['data_type'])
        #         shutil.copy(copy_path, image_path)  # shutil.SameFileError,当同一文件夹拷贝相同文件会出错
        #     if data_info['labelme_file']:
        #         json_path = PathOperate.make_up_path(data_info['labelme_dir'], data_info['labelme_file'], out_dir, None,
        #                                              data_info['data_type'])
        #         # print(json_path)
        #         with open(json_path, "w", encoding='UTF-8') as labelme_fp:  # 以写入模式打开这个文件
        #             json.dump(data_info['labelme_info'], labelme_fp, indent=4, cls=Encoder)  # 从新写入这个文件，把之前的覆盖掉（保存）
        # print('同步耗时')
        # print(time.time() - start_time)
        # ==============================同步写内存数据到磁盘操作============================================================
        # if async_time < start_time:
        #     print('异步写内存数据到磁盘操作,异步耗时短')
        # if async_time > start_time:
        #     print('同步写内存数据到磁盘操作,同步耗时短')

    def visualization(self, out_dir):
        """
        可视化保存图片方法实现（未实现，因为OpenCV调用失败）
        :param out_dir:
        :param replaces:
        """
        global left_top, right_bottom, archive_image_path
        for data_info in self.data_infos:
            image = os.path.join(data_info['image_dir'], data_info['image_file'])

            # image = PathOperate.make_up_path(data_info['image_dir'], data_info['image_file'], None, None)
            # output_image = os.path.join(output_dir, data_info['image_file'])
            img = cv2.imread(str(image))
            # for old, new in replaces.items():
            #     replace_image_path = data_info['image_dir'].replace(old, new)
            #     # replace_json_path = data_info['labelme_dir'].replace(old, new)
            #     # 3. 去掉开头的 斜杠
            #     image_slash = replace_image_path.strip('\\/')
            #     # json_slash = replace_json_path.strip('\\/')
            #     archive_image_path = os.path.join(output_dir, image_slash)
            #     os.makedirs(archive_image_path, exist_ok=True)
            #     # archive_json_path = os.path.join(output_dir, json_slash)
            output_image = image.joinpath(out_dir, image.parent.parent.stem, image.parent.name, data_info['image_file'])
            # output_image = os.path.join(archive_image_path, data_info['image_file'])

            if data_info['labelme_info'] is not None:
                for shape in data_info['labelme_info']['shapes']:
                    # 左上角坐标点
                    left_top = (list(map(int, shape['points'][0])))
                    # 右下角坐标点
                    right_bottom = (list(map(int, shape['points'][1])))
                    # (0, 255, 255) => rgb 颜色,3 => 粗细程度,img => 图片数据,
                    cv2.rectangle(img, left_top, right_bottom, (0, 255, 255), 1)
                    # 保存图片
                    cv2.imwrite(str(output_image), img)

    def __add__(self, other):
        """
        类别筛选合并操作方法
        这合并存在一个问题，就是针对不同合并对象，把固定的第一个对象与传入的其它对象进行比较时，如果传入的对象元素在固定对象中不存在
        则保存输出合并数据时候，针对不在固定对象中的元素会单独创建文件夹保存，导致不为一个整体。如果把筛选的原始对象一起进行合并，可以避免这个问题
        考虑第一个对象有背景的情况，被比较的对象有背景的情况
        :param other:
        :return:
        """
        # 传入第二个对象与第一个对象进行比较
        for index, info in enumerate(other.data_infos):
            # 如果返回的索引不相等，直接把整个元素追加到第一个对象
            if self.find(info['image_file']) is None:
                if info['background']:
                    continue
                self.data_infos.append(info)
            # 如果返回的索引相等，表示为同一张图片，并把第二个对象的shapes元素全部追加到第一个对象
            else:
                other_shapes = list()
                if self.data_infos[self.find(info['image_file'])]['background']:
                    # 当筛选出来的数据，在合并过程中出现了背景，不能够全部跳过。被比较对象如果有值就追加，没有值就跳过
                    if self.data_infos[self.find(info['image_file'])]['labelme_info'] is None:
                        if info.get('labelme_info') is None:
                            continue
                        self.data_infos[self.find(info['image_file'])]['labelme_info'] = info.get('labelme_info')
                        self.data_infos[self.find(info['image_file'])]['background'] = False
                    if not self.data_infos[self.find(info['image_file'])]['labelme_info']['shapes']:
                        try:
                            for shape in info.get('labelme_info').get('shapes'):
                                if shape not in self.data_infos[self.find(info['image_file'])].get('labelme_info').get(
                                        'shapes'):
                                    other_shapes.append(shape)
                            self.data_infos[self.find(info['image_file'])]['background'] = False
                            # 把第二个对象的shapes元素全部追加到第一个对象，并且没有相同的元素
                            self.data_infos[self.find(info['image_file'])].get('labelme_info').get('shapes').extend(
                                other_shapes)
                        except Exception as e:
                            print(e)
                            print(info.get('labelme_info'))
                else:
                    if info.get('labelme_info') is None:
                        continue
                    if not info['labelme_info']['shapes']:
                        continue
                    # 如果俩个对象中的同一张图片下shapes元素全部相同，则跳过追加
                    if self.data_infos[self.find(info['image_file'])].get('labelme_info').get('shapes') \
                            == info.get('labelme_info').get('shapes'):
                        continue
                    # 如果俩个对象中的同一张图片下shapes元素部分相同，则取出第二个对象中不同shape元素
                    for shape in info.get('labelme_info').get('shapes'):
                        if shape not in self.data_infos[self.find(info['image_file'])].get('labelme_info').get(
                                'shapes'):
                            other_shapes.append(shape)
                    # 把第二个对象的shapes元素全部追加到第一个对象，并且没有相同的元素
                    self.data_infos[self.find(info['image_file'])].get('labelme_info').get('shapes').extend(
                        other_shapes)
        return self

    def find(self, image_file):
        """
        查找图片索引，image_file 是否在data_infos里面，如果在就返回下标索引，如果不在返回None用于追加判断
        :param image_file:
        :return:
        """
        for index, info in enumerate(self.data_infos):
            # 返回两个对象相同元素图片的索引
            if info['image_file'] == image_file:
                return index
        return None

    @classmethod
    def merge(cls, datasets):
        merge_dataset = datasets[0]
        # 截取列表元素进行循环
        for dataset in datasets[1:]:
            # 第0个和第一个相加后，赋值给一个变量，然后把这个变量看作一个整体，继续加第三个。
            merge_dataset = merge_dataset + dataset
        merge_dataset.save_labelme()

    def auto_filter(self):
        """
        筛选功能实现,包含label、flags、shape_type
        可自动与自定义
        :return:
        """
        background_element = list()
        self.filter_data['background'] = list()
        self.filter_data['positive'] = list()
        for data_info in self.data_infos:
            if self.only_select_shapes:  # 只筛选标注画框数据（正样本）
                if data_info['labelme_info'] is None:
                    continue
                elif data_info['labelme_info']:
                    if data_info['labelme_info']['shapes']:
                        self.filter_data['positive'].append(data_info)
                        data_info['output_dir'] = os.path.join(data_info['output_dir'], 'positive')
            if self.only_select_empty:  # 只背景筛选（负样本）
                if data_info['labelme_info'] is None:
                    self.filter_data['background'].append(data_info)
                    data_info['output_dir'] = os.path.join(data_info['output_dir'], 'background')
                if data_info['labelme_info']:
                    if not data_info['labelme_info']['shapes']:
                        self.filter_data['background'].append(data_info)
                        data_info['output_dir'] = os.path.join(data_info['output_dir'], 'background')
            if self.shapes_attribute:  # 按shapes的属性筛选，包含label、flags、shape_type
                filter_shapes = list()
                label = list()
                shape_type = list()
                flags = list()
                check_input = ['label', 'flags', 'shape_type']
                if self.shapes_attribute is None or self.shapes_attribute == "''" or \
                        self.shapes_attribute == '' or self.shapes_attribute not in check_input:
                    # 验证是否填写，是否填错
                    assert False, '输入的（--automatic）筛选属性不对'.format(self.shapes_attribute)
                if data_info['labelme_info'] is None:
                    if self.select_empty:
                        background_element.append(data_info)
                    else:
                        continue
                if data_info['labelme_info']:
                    if not data_info['labelme_info']['shapes']:
                        if self.select_empty:
                            background_element.append(data_info)
                        else:
                            continue
                    for shape in reversed(data_info['labelme_info']['shapes']):
                        if self.shapes_attribute == 'label':
                            label.append(shape['label'])
                        if self.shapes_attribute == 'flags':
                            if 'flags' in shape:
                                for key, value in shape['flags'].items():
                                    if shape['flags'][key] is True:
                                        flags.append(key)
                        if self.shapes_attribute == 'shape_type':
                            shape_type.append(shape['shape_type'])
                if self.filter_label is None and self.shapes_attribute == 'label':
                    # 自动筛选时，筛选元素集汇总。每一张图片的类别与整个对象中多个图片的类别取相同的部分
                    filter_shapes = set(label)
                elif self.filter_label and self.shapes_attribute == 'label':
                    # 指定元素筛选时，筛选元素集汇总
                    filter_shapes = set(label) & set(self.filter_label)
                if self.filter_flags is None and self.shapes_attribute == 'flags':
                    filter_shapes = set(flags)
                elif self.filter_flags and self.shapes_attribute == 'flags':
                    filter_shapes = set(flags) & set(self.filter_flags)
                if self.filter_shape_type is None and self.shapes_attribute == 'shape_type':
                    filter_shapes = set(shape_type)
                elif self.filter_shape_type and self.shapes_attribute == 'shape_type':
                    filter_shapes = set(shape_type) & set(self.filter_shape_type)
                for label_key in filter_shapes:
                    label_flags = deepcopy(data_info)
                    # 同一个对象中同一张图片的相同类别筛选
                    if label_key not in self.filter_data:
                        self.filter_data[label_key] = []
                        label_flags['labelme_info']['shapes'] = self.data_info_shapes(data_info, label_key,
                                                                                      self.filter_combin)
                        self.filter_data[label_key].append(label_flags)
                        # if self.only_select_shapes is False:
                        label_flags['output_dir'] = os.path.join(label_flags['output_dir'], label_key)
                    # 同一个对象，不同图片的相同类别筛选
                    if not (self.filter_data[label_key][-1]['image_dir'] == data_info['image_dir']
                            and self.filter_data[label_key][-1]['image_file'] == data_info['image_file']):
                        label_flags['labelme_info']['shapes'] = self.data_info_shapes(data_info, label_key,
                                                                                      self.filter_combin)
                        self.filter_data[label_key].append(label_flags)
                        # if self.only_select_shapes is False:
                        label_flags['output_dir'] = os.path.join(label_flags['output_dir'], label_key)
        if self.select_empty:  # 筛选label时，把背景元素追加到每个类别
            for key in self.filter_data.keys():
                for bk in deepcopy(background_element):  # 同一个对象反复迭代修改值，必须深度拷贝
                    # 根据键修改值，需要深度拷贝一份，否则相同的键为同时一起修改
                    bk['output_dir'] = os.path.join(self.output_dir, key)
                    self.filter_data[key].append(bk)
            del self.filter_data['background']
            del self.filter_data['positive']
        return self

    @staticmethod
    def data_info_shapes(data_info, label_key, filter_combin):
        """
        shapes属性对象处理
        组合筛选预想思路，以label为准，查找是否存在flags为真的，然后追加shape元素
        reversed()逆序遍历方便删除列表元素后继续迭代,并转列表为可迭代对象，减少内存消耗
        :param filter_combin:
        :param data_info:
        :param label_key:
        :return:
        """
        filter_result = list()
        for shape in reversed(data_info['labelme_info']['shapes']):
            # if shape['label'] == label_key and filter_combin:
            #     for key, value in shape['flags'].items():
            #         if shape['flags'][key] is True:
            #             filter_result.append(shape)
            if shape['label'] == label_key:
                filter_result.append(shape)
            if shape['shape_type'] == label_key:
                filter_result.append(shape)
            for key, value in shape['flags'].items():
                if shape['flags'][key] is True:
                    if key == label_key:
                        filter_result.append(shape)
        return filter_result

    def del_label(self, label):
        """
        指定label类别进行删除操作
        :param label:
        """
        for data_info in self.data_infos:
            if data_info['labelme_info'] is None:
                continue
            elif data_info['labelme_info']:
                for shape in reversed(data_info['labelme_info']['shapes']):
                    # print(shape)
                    if shape['label'] in label:
                        data_info.get('labelme_info').get('shapes').remove(shape)
